require 'date'
require 'parsedate'

=begin
= timex

This library extends Time class:
* date string parsing.
* stringization for RFC 2822 and ISO 8601.
* leapseconds and localtime offset information.

This library has following assumptions:
* An offset from UTC to localtime is a multiple of 60 seconds.
* Number of leapseconds are less than 60.
While it is true now, in 2001, it will be invalidated in some future.
(20 years? 100 years?)
I hope that POSIX will have better API for handling time which enables
abolish these assumptions.
=end

class Time
  class << Time

=begin
== class methods

--- Time.parse(date, cyear=false)
    parses ((|date|)) using ParseDate.parsedate and convert it to Time object.
    ((|cyear|)) is passed to ParseDate.parsedate.

    When ParseDate.parsedate fails to extract some components of
    a broken-down time, (({Time.now})) is used for them.

    Since there are various timezone abbreviations in the world,
    conflicts exist between some of them.
    For example, CST is used as
    -06:00 in America/Chicago,
    -05:00 in America/Havana,
    +08:00 in Asia/Harbin,
    +09:30 in Australia/Darwin,
    +10:30 in Australia/Adelaide,
    etc.
    So, it is impossible to support all of them.
    
    This method supports timezone abbreviations described in RFC 822 and
    system timezone.
    The system timezone is taken from
    (({Time.local(year, 1, 1).zone})) and
    (({Time.local(year, 7, 1).zone})).
    If the extracted timezone abbreviation doesn't match them,
    it is ignored and broken-down time is assumed as represented in localtime.
    If it matches RFC 822 and system timezone, RFC 822 is prefered.
=end
    ZoneOffset = {
      'UTC' => 0,
      # ISO 8601
      'Z' => 0,
      # RFC 822
      'UT' => 0, 'GMT' => 0,
      'EST' => -5, 'EDT' => -4,
      'CST' => -6, 'CDT' => -5,
      'MST' => -7, 'MDT' => -6,
      'PST' => -8, 'PDT' => -7,
    }
    def parse(date, cyear=false)
      year, mon, day, hour, min, sec, zone, _ = ParseDate.parsedate(date, cyear)

      t = nil
      year = (t ||= Time.now).year unless year
      mon = (t ||= Time.now).mon unless mon
      day = (t ||= Time.now).day unless day
      hour = (t ||= Time.now).hour unless hour
      min = (t ||= Time.now).min unless min
      sec = (t ||= Time.now).day unless sec

      off = nil
      if zone
	zone.upcase!
	if /\A([-+]\d\d):?(\d\d)\z/ =~ zone
	  off = ($1.to_i * 60 + $2.to_i) * 60
	elsif /\A[-+]\d\d\z/ =~ zone
	  off = zone.to_i * 3600
	elsif ZoneOffset.include?(zone)
	  off = ZoneOffset[zone.upcase] * 3600
	elsif (t = Time.local(year, 1, 1)).zone.upcase == zone
	  off = t.utcoff
	elsif (t = Time.local(year, 7, 1)).zone.upcase == zone
	  off = t.utcoff
	end
      end

      if off
        t = Time.utc(year, mon, day, hour, min, sec) - off
	t.localtime if off != 0
	t
      else
        Time.local(year, mon, day, hour, min, sec)
      end
    end
  end

  EpochJD = Date.new3(1970, 1, 1).jd
  def _utcoff_leapsecs
    days = Date.new3(year, month, day).jd - EpochJD
    diff = tv_sec - (sec + 60 * (min + 60 * (hour + 24 * days)))
    utcoff, leapsecs = diff.divmod(60)
    utcoff *= -60
    return utcoff, leapsecs
  end

=begin
== methods

--- Time#utcoff
--- Time#gmtoff
    returns difference to UTC in seconds.

    Note that it returns 0 if self is UTC mode.
=end
  def utcoff
    _utcoff_leapsecs[0]
  end
  alias gmtoff utcoff

=begin
--- Time#leapsecs
    returns number of leapseconds from Unix epoch.

    Note that this method returns non-zero only if
    your system has leapseconds information.
    Since POSIX defines that leapseconds are not handled,
    most OSs doesn't handle leapseconds by default.
    But many OSs can handle leapseconds fairly easily in theory
    (but not well tested, though).
    For example, GNU/Linux (and various OSs) uses zoneinfo database to
    store leapseconds information and
    some GNU/Linux distributions has zoneinfo database which contains
    leapseconds information under /usr/share/zoneinfo/right.
    So, under such Linux, setting TZ environment variable is what you need
    to use leapseconds.
=end
  def leapsecs
    _utcoff_leapsecs[1]
  end

=begin
--- Time#rfc2822
--- Time#rfc822
    returns a string which represents the time in RFC 2822 date format.
    The format is identical to RFC 822 date format modified by RFC 1123.

    If self is UTC mode, the result is also complient with
    HTTP-date (rfc1123-date) defined by RFC 2616 which is restricted version of
    the date format defined by RFC 822 and RFC 1123.
=end
  RFC2822_DAY_NAME = [
    'Sun', 'Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat'
  ]
  RFC2822_MONTH_NAME = [
    'Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun',
    'Jul', 'Aug', 'Sep', 'Oct', 'Nov', 'Dec'
  ]
  def rfc2822
    sprintf('%s, %02d %s %d %02d:%02d:%02d ',
      RFC2822_DAY_NAME[wday],
      day, RFC2822_MONTH_NAME[mon-1], year,
      hour, min, sec) +
    (utc? ? 'GMT' : sprintf('%+03d%02d', *(utcoff / 60).divmod(60)))
  end
  alias rfc822 rfc2822

=begin
--- Time#iso8601
    returns a string which represents the time in ISO 8601 format.

    However ISO 8601 defines various format for date/time,
    the method supports single format: YYYY-MM-DDThh:mm:ssTZD
=end
  def iso8601
    sprintf('%d-%02d-%02dT%02d:%02d:%02d',
      year, mon, day, hour, min, sec) +
    (utc? ? 'Z' : sprintf('%+03d:%02d', *(utcoff / 60).divmod(60)))
  end
end
